%% Per Mildner 1995
%% As of 3PL3 trace now calls user:debugger_write_goal_hook/7 if available
%% this is how we hook in to give Emacs the necessary information
%% Changes, most recent first
%% 970503 . Adding support for XEmacs.
%%          Call emacs_breakable_lines to tell emacs about valid lines
%%          for breakpoints.
%%          Consider hooking into compile/consult with a
%%          term_expansion for end_of_file.
%%          Rensa breakable lines etc tminstne vid
%%          emacs_consult/_compile

%%
%% 960930 . output the "remove temporary break" command BEFORE the
%%          ordinary src pos output. This way the questionmark will
%%          always come immediately after the ordinary output.
%% 960923 . wrap output in with_unbuffered_user_error instead of using
%%          kludge_open et al. explicitly
%% 960912 . emacs_temporary_break(File, Line) sets a break-point that
%%          will be removed as soon as it appears.
%%          Really needs some feedback to emacs if breakpoint could
%%          not be set.
%%
%% 960902
%%        . now uses user:Path(Line) instead of '$ pos'(Line,Path)
%%        . Check if we're running under emacs before doing fancy output
%%
%% 960811 . writeq is now off by default. 
%%
%% 960525
%%        . prefixed the prefix_file_info with ~N. Unfortunately no effect
%%          The new regexps require specially processed output to
%%          begin at a newline. This is not the case when doing, eg,
%%          C-C c (consult) from the debugger. the idea then is to
%%          ensure this by prefixing the file_info with
%%          ~N. Unfortunately format seems to have the wrong idea
%%          about whether there is a newline there or not (or it might
%%          be that the emacs stuff has already deleted a newline).
%%
%% 960222
%%        . Changed from source_debug:pos/2 to user:'$ pos'/2 (speed factor 2 for nrev!).
%% 960115
%%        . Fixed breakpoints. It broke when we wrapped this in a module.
%% 960105
%%        . Added do_command, should add a version that gets an earlier
%%          goal from emacs too.
%%        . emacs_trace_enable(Old, New), on/off (default on) to
%%          disable the fancy debugger output.
%%        . emacs_writeq_enable(Old, New), on/off (default off as of 960811) to
%%          disable the output of writeq versions of all args.
%%          This is useful if your args are circular or very large.


%% ToDo:
%% . Rename all things called from emacs consistently
%% . We should let the ordinary code do the writing after we reach
%%   uninteresting levels.


%% NOTE! module name occurs explicitly in add_line_condition/2
%%       Beware of dynamic preds passed to non-meta declared
%%       prolog preds!
:- module(emacs, [emacs_trace_enable/2, emacs_writeq_enable/2]).

:- use_module(library(lists), [memberchk/2,member/2]).
:- use_module(library(terms), [acyclic_term/1]).
:- use_module(library(system), [environ/2]).

:- meta_predicate with_unbuffered_user_error(:).

%% don't! will be catenated into the same file.
%% :- ensure_loaded(kludge_stream).                % buffered write hack

:- dynamic trace_hook_enabled_flag/0.
:- dynamic break_line_condition/2.

:- meta_predicate without_user_error(:).
/*
%% user:term_expansion/2 is defined at end of file
term_expansion_hook :-
   term_expansion_hook1,
   !, fail.
*/

/* 970613 Turned out that it is better to do this in a portray_message hook
%% Will be called on end_of_file when compiling, consulting, konsulting.
term_expansion_hook1 :- source_debug:konsult_context(file, File), !,
   konsult_hook(File).
term_expansion_hook1 :- prolog_load_context(file, File), !,
   non_konsult_hook(File).
*/

%% konsult_hook(Path) :-
%%    %% must be done a little later (in act_finish).
%%    %% emacs_file_breakpoint_info1(Path)
%%    true.


%% Called for consulted/compiled/loaded files from portray_message
non_konsult_hook(Path, _Module) :-
   current_predicate(remove_source_debug_info, source_debug:remove_source_debug_info(_)),!,
   source_debug:remove_source_debug_info(Path),
   emacs_file_breakpoint_info1(Path).
   



without_user_error(Goal) :-
   prolog:'$persistent_null_stream'(NullStream),% could use open_null_stream
   prolog_flag(user_error, User_Error),
   flush_output(User_Error),
   (
       on_exception(E,
                    ( prolog_flag(user_error, _, NullStream),
                      call(Goal)
                    ),
                    ( prolog_flag(user_error, _, User_Error),
                      raise_exception(E)
                    )),
       !, 
       prolog_flag(user_error, _, User_Error)
   ; prolog_flag(user_error, _, User_Error),
     !,
     fail
   ).

                  
              

:- meta_predicate preserve_debug(:).

preserve_debug(Goal) :-
   prolog_flag(debugging, Old),
   (
       on_exception(E,
                    Goal,
                    ( prolog_flag(debugging, _, Old),
                      raise_exception(E)
                    )),
       !,
       prolog_flag(debugging, _, Old)
   ; prolog_flag(debugging, _, Old),
     !,
     fail
   ).


%% default enabled
trace_hook_enabled_flag.


trace_hook_enabled :-
   trace_hook_enabled_flag,
   running_in_emacs.


running_in_emacs :-
   %% 970820 EMACS is t even for subprocesses such as xterm!
   %% environ('EMACS', t),
   environ('TERM', 'emacs'),
   !.                                           % probably not needed


emacs_trace_enable(Old, New) :-
   emacs_enable(Old, New, trace_hook_enabled_flag).


emacs_enable(Old, New, Flag) :-
   ( call(Flag) ->
     Old = on
   ; Old = off
   ),
   ( New == on ->
     retractall(Flag),
     assert(Flag)
   ; New == off ->
     retractall(Flag)
   ).

:- dynamic trace_hook_writeq_enabled_flag/0.

%% 960811 now off by default
%% trace_hook_writeq_enabled_flag.

emacs_writeq_enable(Old, New) :-
   emacs_enable(Old, New, trace_hook_writeq_enabled_flag).

%% As of SICStus 3 PL3 this is what is called
user:debugger_write_goal_hook(A,B,C,D,E,F,G) :-
   trace_hook_enabled,
   write_goal_hook(A,B,C,D,E,F,G).

%% called from trace.pl
%% If this ever gets out to the public then we should reconsider the
%% arguments to the hook
write_goal_hook(_A,_B,_C,_D,_E,_F,_G) :-
   write_goal_skip_flag, !,
   write_goal_skip(off),
   format(user_error, '10Lignore10R', []).

write_goal_hook(_T, _Goal, _Xs, B, _D, _Port, _Module) :- B == 1,
   %% Do this when starting a debug session.
   dump_all_breakpoint_info,
   fail.

write_goal_hook(_T, Goal, _Xs, _B, _D, _Port, _Module) :-
   pos_marker_goal(Goal,Pos,File), !,           % 960826 don't do direct match
   do_pos_marker1(Pos,File).                    % 970509 moved out
%% 960912 gone
%% %% 960222 hide the new profiling goal
%% write_goal_hook(_T, source_debug:pos_profile(Pos,File), _Xs, _B, _D, _Port, _Module) :- !,
%%    format(user_error, '9L~w9M~w9R', [Pos,File]).
%% write_goal_hook(_T, source_debug:pos(Pos,File), _Xs, _B, _D, _Port, _Module) :- !,
%%   format(user_error, '9L~w9M~w9R', [Pos,File]).
write_goal_hook(T, X, Xs, B, D, Port, Module) :-
   with_unbuffered_user_error(write_goal_hook1(T, X, Xs, B, D, Port, Module)).


write_goal_hook1(T, X, Xs, B, D, Port, Module) :-
   prolog:port_info(Port, Pport),
   prolog:port_info(Xs, X, Module, Mark, S, []),
   
   %% While we're hacking we might as well rigth-justify
   %% the depth and invocation number
   format(user_error, '0L~a7L~|~t~w~3+7R ~|~t~w~3+8L~a8R', 
          [Mark,B,D,Pport]),
   %% format(user_error, '~a~w  ~w~a', [Mark,B,D,Pport]),
   prolog:errdisplay_list(S),
   ( Xs = [] ->
       dbg_write_goal(T, X)
   ; otherwise ->
       %% some subterm, give up
       prolog:write_goal(T, X)
   ),
   %% 960216 Akk, not do user_error:
   %% prolog:'$display'('0R').
   write(user_error, '0R'),
   flush_output(user_error).                    % for good measure

%% 970509 called from write_goal_hook1
do_pos_marker1(Pos,File) :-
   pos_marker_seen(Pos, File),
   format(user_error, '9L~w9M~w9R', [Pos,File]).


do_pos_marker(Pos,File) :-
   pos_marker_seen(Pos, File),
   %% the newline should prevent emacs from sending a creep command in
   %% response to the pos info
   format(user_error, '9L~w9M~w9R~n', [Pos,File]).


%% 970509 called from break point code
do_pos_marker_for_break(Pos,File) :-
   do_pos_marker(Pos,File),
   %% nl(user_error), % prepend trace message with newline
   %% trace

   %% settgin trace with prolog_flag does not write message!
   %% without_user_error(trace).
   prolog_flag(debugging, _, trace).





pos_marker_goal(user:Goal,Pos,File) :-
   source_debug:'pos marker goal'(Goal), !,
   Goal =..[File,Pos].

pos_marker_goal(Goal,Pos,File) :-
   source_debug:'pos marker goal'(Goal), !,
   Goal =..[File,Pos].

pos_marker_seen(Pos, File) :-
   remove_temporary_break(File, Pos).




/* to 960902
%% 960826 this is to make it easier to experiment with other formats
pos_marker_goal('$ pos'(Pos0,File0),Pos,File) :- !,
   Pos = Pos0,
   File = File0.

pos_marker_goal(user:'$ pos'(Pos0,File0),Pos,File) :- !,
   Pos = Pos0,
   File = File0.
*/



dbg_write_goal(T, X) :- var(X), !,              % ? should not happen
   prolog:write_goal(T, X).
dbg_write_goal(0'd, Goal) :- prolog:display1(user_error, Goal).% display
dbg_write_goal(0'p, Goal) :- dbg_errprint(Goal).% print
dbg_write_goal(p(I), Goal) :-                   % print restricted
   %%'$printdepth'(Old, I),
   dbg_errprint(Goal, I),
   %% '$printdepth'(_, Old).
   true.
dbg_write_goal(0'w, Goal) :-
   dbg_writeq(Goal).                            % write





%% support stuff called from emacs
emacs_toggle_leash(What) :-
   write_goal_skip(on),
   prolog:is_leashed([call,exit,redo,fail,exception], L),
   ( dbg_delete(L, What, L1) ->
       leash(L1)
   ; otherwise ->
       leash([What|L])
   ).

dbg_delete([X|Xs], X1, Xs1) :- X==X1, !,
   Xs1 = Xs.
dbg_delete([X|Xs], X1, [X|Xs1]) :-
   dbg_delete(Xs, X1, Xs1).


dbg_member(X, [X|_Xs]).

dbg_member(X, [_|Xs]) :-
   dbg_member(X, Xs).

%% used for conditional breakpoints
dbg_install_condition(Goal, Port, Cond, Module) :-
   prolog:spypoint_condition(Module:Goal, Port, Cond),
   write_goal_skip(on).

%% used to compile an interpreted goal
%% The module juggling is crazy but seems necessary
%% If already compiled then no harm will be done
dbg_compile_pred(Module, F, A, TempPath) :-
   %% TempPath = 'tmp_sicstus_dbg_compile.pl',
   
   functor(X, F, A),
   prolog:'$source file'(X, Module, Path),

   prolog_flag(typein_module, OldModule, prolog),

   tell(TempPath),
   listing(Module:(F/A)),
   told,
   module(Module),
   prolog_flag(redefine_warnings, Old, off),
   compile(TempPath),
   prolog_flag(redefine_warnings, _, Old),
   prolog_flag(typein_module, _, OldModule),
        %% restore the original source path
   retractall(prolog:'$source file'(X, Module, _)),
   asserta(prolog:'$source file'(X, Module, Path)).




%% Consider special write_atom so that any maybe_space is done
%% before any wrapper
%% consider a new write_args/2 so that write_goal puts an extra space
%% after comma before args too.

dbg_write_out_top(X, Style) :-
   functor(X, F0, A0),                                                                     % should always succeed
   ( X = Module:Goal ->
       atom(Module),
       functor(Goal, F, A),
       Depth = -4
   ; otherwise ->
       Goal = X,
       F = F0, A = A0,
       Depth = -2
   ), 
   F/A \== '$VAR'/1,
   F/A \== '{}'/1,
   F \== '.', F \== ',', !,
   Prio = 1200,

   %% We use Depth = -2 to flag that we wan't the special wrappers
   %% We use Depth = -4 to mark that we wan't wrappers and that
   %% there is a module
   %% These will be decremented by one before we actually look at them

   write_out(X, Style, Prio, 0, Depth, '(', 2'100, _).

dbg_write_out_top(X, Style) :-
   %% The goal was strange, back off and let the original code do it.
   prolog:write_out_top(X, Style, 1200).


write_out(Term, _, _, _, _, _, Ci, 2'000) :-            % Original code
        var(Term), !,
        prolog:maybe_space(Ci, 2'000),
        prolog:'$write'(Term).
write_out('$VAR'(N), Style, _, _, _, _, Ci, Co) :- % Original code
        Style = style(_,_,_,true,_,_,_),
        prolog:write_VAR(N, Ci, Co), !.
write_out(_, Style, _, _, Depth, _, Ci, 2'010) :- % Original code
        Style = style(_,_,_,_,_,_,MaxDepth),
        Depth >= MaxDepth, !,
        prolog:maybe_space(Ci, 2'010),
        prolog:'$display'(...).
write_out(Term, Style, _, _, Depth, _, _, 2'000) :-
        
        %% [PM] Don't portray specially formatted goals
        Depth > -2,
        
        Style = style(_,_,true,_,_,_,_),
        (   \+prolog:call_user_def(portray(Term), user) ->
            fail                 % portray might bind variables
        ;   true
        ), !.
write_out(Atom, Style, _, PrePrio, Depth, Lpar, _, 2'100) :-
        atom(Atom),
        prolog:current_prefixop(Atom, P, _),
        P =< PrePrio, !,
        prolog:'$display'(Lpar),
        Style = style(Quoted,_,_,_,_,_,_),

        maybe_wrap(Depth, '('),                                                         % [PM]
        prolog:write_atom(Quoted, Atom, 2'100, _),
        maybe_wrap(Depth, ')'),                                                         % [PM]

        put(0')).
write_out(Atom, Style, _, _, Depth, _, Ci, Co) :-
        atom(Atom), !,
        Style = style(Quoted,_,_,_,_,_,_),
        maybe_wrap(Depth, '('),                                                         % [PM]
        prolog:write_atom(Quoted, Atom, Ci, Co),
        maybe_wrap(Depth, ')').                                                         % [PM]

write_out(N, _, _, _, _, _, Ci, 2'000) :-                       % Original code
        integer(N), !,
        (   N < 0 -> prolog:maybe_space(Ci, 2'010)
        ;   prolog:maybe_space(Ci, 2'000)
        ),
        prolog:'$write'(N).
write_out(N, _, _, _, _, _, Ci, 2'000) :-                       % Original code
        float(N), !,
        (   N==0.0 -> prolog:maybe_space(Ci, 2'000)     % 0.0 but not -0.0
        ;   N < 2.0*N -> prolog:maybe_space(Ci, 2'000)  % >0.0
        ;   prolog:maybe_space(Ci, 2'010)               % <0.0/+inf/-inf/+nan/-nan
        ),
        prolog:'$write'(N).
write_out(Term, Style, _, _, Depth, _, Ci, 2'100) :-
        Style = style(Quoted,true,_,_,_,_,_),
        functor(Term, Atom, Arity), !,

        maybe_wrap(Depth, '('),                                                         % [PM]
        prolog:write_atom(Quoted, Atom, Ci, _),
        maybe_wrap(Depth, ')'),                                                         % [PM]

        Depth1 is Depth+1,
        write_args(0, Arity, Term, Style, Depth1).
write_out({Term}, Style, _, _, Depth, _, _, 2'100) :- !,        % Original code
        put(0'{),
        Depth1 is Depth+1,
        prolog:write_out(Term, Style, 1200, 0, Depth1, '(', 2'100, _),
        put(0'}).
write_out([Head|Tail], Style, _, _, Depth, _, _, 2'100) :- !, % Original code
        put(0'[),
        Depth1 is Depth+1,
        prolog:write_out(Head, Style, 999, 0, Depth1, '(', 2'100, _),
        prolog:write_tail(Tail, Style, Depth1).
write_out((A,B), Style, Prio, _, Depth, Lpar, Ci, Co) :- !,     % Original code
        %  This clause stops writeq quoting commas.
        Depth1 is Depth+1,
        prolog:maybe_paren(1000, Prio, Lpar, Lpar1, Ci, C1),
        write_out(A, Style, 999, 0, Depth1, Lpar1, C1, _),
        put(0',),
        prolog:write_out(B, Style, 1000, 1000, Depth1, '(', 2'100, C2),
        prolog:maybe_paren(1000, Prio, C2, Co).
write_out(Term, Style, Prio, PrePrio, Depth, Lpar, Ci, Co) :-
        functor(Term, F, N),
        Depth1 is Depth+1,
        write_out(N, F, Term, Style, Prio, PrePrio, Depth1, Lpar, Ci, Co).

%% [PM] This is called with Depth either:
%%      0 -- This is from portray_clause and friends
%%     -1 -- New! Flags that we should wrap the goal arguments
%%     -3 -- New! Flags that we should wrap module too.
write_out(1, F, Term, Style, Prio, _, Depth0, Lpar, Ci, Co) :-
        prolog:current_postfixop(F, P, O), !,
        (prolog:current_infixop(F, _, _, _) -> O1=1200; O1=O),
        prolog:maybe_paren(O1, Prio, Lpar, Lpar1, Ci, C1),
        arg(1, Term, A),

        maybe_wrap_arg(Depth0, '(', Depth),                             % [PM]
        write_out(A, Style, P, 1200, Depth, Lpar1, C1, C2),
        maybe_wrap_arg(Depth0, ')'),                                            % [PM]

        Style = style(Quoted,_,_,_,_,_,_),

        maybe_wrap(Depth0, '('),                                                        % [PM]
        prolog:write_atom(Quoted, F, C2, C3),
        maybe_wrap(Depth0, ')'),                                                        % [PM]

        prolog:maybe_paren(O1, Prio, C3, Co).
write_out(1, F, Term, Style, Prio, PrePrio, Depth0, Lpar, Ci, Co) :-
        F \== -, F \== +,
        prolog:current_prefixop(F, O, P), !,
        (PrePrio=1200 -> O1 is P+1; O1=O),      % for "fy X yf" etc. cases
        prolog:maybe_paren(O1, Prio, Lpar, _, Ci, C1),
        Style = style(Quoted,_,_,_,_,_,_),
        maybe_wrap(Depth0, '('),                                                        % [PM]
        prolog:write_atom(Quoted, F, C1, C2),
        maybe_wrap(Depth0, ')'),                                                        % [PM]
        arg(1, Term, A),

        maybe_wrap_arg(Depth0, '(', Depth),                             % [PM]
        write_out(A, Style, P, P, Depth, ' (', C2, C3),
        maybe_wrap_arg(Depth0, ')'),                                            % [PM]

        prolog:maybe_paren(O1, Prio, C3, Co).
write_out(2, F, Term, Style, Prio, PrePrio, Depth0, Lpar, Ci, Co) :-
        prolog:current_infixop(F, P, O, Q), !,
        (PrePrio=1200 -> O1 is Q+1; O1=O),      % for "U xfy X yf" etc. cases
        prolog:maybe_paren(O1, Prio, Lpar, Lpar1, Ci, C1),
        arg(1, Term, A),

        %% [PM] Special case for Module:Goal
        %% if Depth is -3 then Arg1 is the Module atom, it should
        %% be wrapped but the Depth will not matter.
        %% Arg 2 should not be wrapped here and should be written
        %% with depth -1 (it is the goal).

        ( Depth0 = -3 ->
                 DepthA0 = -3,
                 DepthB0 = -2                                                                           % tells maybe_wrpa_arg to no-op
        ; DepthA0 = Depth0,
          DepthB0 = Depth0
   ),

        maybe_wrap_arg(DepthA0, '(', DepthA),                   % [PM]
        write_out(A, Style, P, 1200, DepthA, Lpar1, C1, C2),
        maybe_wrap_arg(DepthA0, ')'),                                           % [PM]

        Style = style(Quoted,_,_,_,_,_,_),

        maybe_wrap(Depth0, '('),                                                        % [PM]
        prolog:write_atom(Quoted, F, C2, C3),
        maybe_wrap(Depth0, ')'),                                                        % [PM]

        arg(2, Term, B),

        maybe_wrap_arg(DepthB0, '(', DepthB),                   % [PM]
        write_out(B, Style, Q, Q, DepthB, '(', C3, C4),
        maybe_wrap_arg(DepthB0, ')'),                                           % [PM]

        prolog:maybe_paren(O1, Prio, C4, Co).
write_out(N, F, Term, Style, _, _, Depth0, _, Ci, 2'100) :-
        Style = style(Quoted,_,_,_,_,_,_),

        maybe_wrap(Depth0, '('),                                                        % [PM]
        prolog:write_atom(Quoted, F, Ci, _),
        maybe_wrap(Depth0, ')'),                                                        % [PM]

        write_args(0, N, Term, Style, Depth0).


%   write_args(DoneSoFar, Arity, Term, Style, Depth)
%   writes the remaining arguments of a Term with Arity arguments
%   all told in Style, given that DoneSoFar have already been written.

write_args(N, N, _, _, _) :- !,
        put(0')).
write_args(I, _, _, Style, Depth) :-
        Style = style(_,_,_,_,_,_,MaxDepth),
        Depth >= MaxDepth,!,
        prolog:write_args(I,Depth),
        prolog:'$display'(...),
        put(0')).
write_args(I, N, Term, Style, Depth0) :-
        prolog:write_args(I, Depth0),
        J is I+1,
        arg(J, Term, A),

        maybe_wrap_arg(Depth0, '(', Depth),                             % [PM]
        write_out(A, Style, 999, 0, Depth, '(', 2'100, _),
        maybe_wrap_arg(Depth0, ')'),                                            % [PM]

        write_args(J, N, Term, Style, Depth0).



maybe_wrap(-1, LR, Depth) :- !, % pred name
        Depth = 0,
        ( LR = '(' ->
                 prolog:'$display'('2L')
        ; LR = ')' ->
       prolog:'$display'('2R')
        ).

maybe_wrap(-2, LR, Depth) :- !, % pred name no args (directly from write_out/8)
        Depth = 0,
        ( LR = '(' ->
                 prolog:'$display'('2L')
        ; LR = ')' ->
       prolog:'$display'('2R')
        ).

maybe_wrap(-3, _LR, Depth) :- !,                                                % module colon, no special wrapping
        Depth = 0.                                                                                              % not actually used

maybe_wrap(Depth, _LR, Depth).


maybe_wrap(Depth, LR) :-
        maybe_wrap(Depth, LR, _Depth).


maybe_wrap_arg(-1, LR, Depth) :- !, % toplevel arg
        Depth = 0,
        ( LR = '(' ->
                 prolog:'$display'('3L')
        ; LR = ')' ->
       prolog:'$display'('3R')
        ).

maybe_wrap_arg(-2, _LR, Depth) :- !,                            % Goal after module:, no wrap
        Depth = -2.                                                                                             % depth for goal

maybe_wrap_arg(-3, LR, Depth) :- !, % module name
        Depth = 0,
        ( LR = '(' ->
                 prolog:'$display'('4L')
        ; LR = ')' ->
       prolog:'$display'('4R')
        ).

maybe_wrap_arg(Depth, _LR, Depth).


maybe_wrap_arg(Depth, LR) :-
        maybe_wrap_arg(Depth, LR, _Depth).



/*
write_goal(0'd, Goal) :- display1(user_error, Goal). % display
write_goal(0'p, Goal) :- errprint(Goal).        % print
write_goal(p(I), Goal) :-                       % print restricted
        '$printdepth'(Old, I),
        errprint(Goal),
        '$printdepth'(_, Old).
write_goal(0'w, Goal) :- writeq(user_error, Goal). % write

errprint(X) :-
        '$printdepth'(D, D),
        '$unknown'(Old, fail),
        '$output'(Curr, user_error),
        Style = style(false,false,true,true,false,false,D),
        write_out_top(X, Style, 1200),
        '$output'(_, Curr),
        '$unknown'(_, Old).
*/




dbg_writeq(X) :-
        Quote = writeq,
        Depth = 16777216,                       % posinf
        dbg_errprint(X, Quote, Depth).


% if using PL6, remove the $ sign
dbg_errprint(X) :-
%        prolog:'$printdepth'(Depth, Depth),
        prolog:'printdepth'(Depth, Depth),
        dbg_errprint(X, Depth).


dbg_errprint(X, Depth) :-
        Quote = print,
        dbg_errprint(X, Quote, Depth).


dbg_errprint(X, _Quote, Depth) :-

        %% prolog:'$printdepth'(D, D),
        prolog:'$unknown'(Old, fail),
        prolog:'$output'(Curr, user_error),
        %% writeq Style = style(true,false,false,true,false,false,16777216),
        %% write Style = style(false,false,false,true,false,false,16777216),
        %% print Style = style(false,false,true,true,false,false,16777216),
        WriteqStyle = style(true,false,false,true,false,false,16777216),
        PrintStyle = style(false,false,true,true,false,false,Depth),
        %%      ( Quote = writeq ->
        %%               WriteqStyle = style(true,false,false,true,false,false,Depth)
        %%      ; Quote = print ->
        %%          PrintStyle = style(false,false,true,true,false,false,Depth)
        %%      ),

        %% prolog:write_out_top(X, Style, 1200),
        
        %% kludge_stream_open(user_error, Cookie),
        on_exception(Err,
           (
           prolog:'$display'('1L'),
           ( (acyclic_term(X), trace_hook_writeq_enabled_flag) ->
               dbg_write_out_top(X, WriteqStyle)
           ; otherwise ->
               true),
            prolog:'$display'('1M'),
            dbg_write_out_top(X, PrintStyle),
            
            prolog:get_module(X, Goal, Module), % get its module or var
            prolog:home_module(Goal, Module, HomeModule),

            %% source file (backtrack to right Module)
            ( prolog:'$source file'(X, HomeModule, Path) ->
                prolog:'$display'('5L'),
                prolog:'$display'(Path),
                prolog:'$display'('5R')
            ; otherwise ->
                true
            ),
            
            prolog:'$display'('6L'),
            prolog:'$display'(HomeModule),
            prolog:'$display'('6R'),

            prolog:'$display'('1R')
           ), ( %% kludge_stream_close(Cookie),
                prolog:'$output'(_, Curr),
                prolog:'$unknown'(_, Old),
                raise_exception(Err) )),
        %% kludge_stream_close(Cookie),

        prolog:'$output'(_, Curr),
        prolog:'$unknown'(_, Old).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% asserted if write_goal_hook should justs print a special ignore line.
:- dynamic write_goal_skip_flag/0.

write_goal_skip_flag :- fail.

'executing_in_@_debugger_command' :-
   %% 961003 zero if at top-loop or in break.
   %%        > 0 in @-command in debugger.
   %% The problem was consult-command etc from emacs sent while at top-level.
   prolog:'$debugger_invocation'(BoxNo),
          ( integer(BoxNo) ->                   % 3PL3
              BoxNo > 0
          ; get_mutable(X,BoxNo) ->     % 3PL5
              X > 0
          ).

write_goal_skip(on) :-
   ( 'executing_in_@_debugger_command' ->
       retractall(write_goal_skip_flag),
       assert(write_goal_skip_flag)
   ; true ).

write_goal_skip(off) :-
   retractall(write_goal_skip_flag).





%% The argument order is significant. It is an attempt to benefit from
%% first argument indexing in interpreted code. There is likely to be
%% more break-points for a particular file than for a particular line.
%% 960912 no dummy-def should be needed
%% break_line_condition(0,dummy).


/* from Bips/trace.pl 3PL3. note that install_spypoint removes all existing conditions.
spypoint_condition(Goal0, Port, Cond) :-
   get_module(Goal0, Goal, Module),
   check_module(Module, Goal, spypoint_condition(Goal0,Port,Cond), 1),
   install_spypoint(Goal, Module),
   asserta_internal('$spypoint_condition'(Goal,Module,Port,Cond), _).
*/

   
add_line_condition(Line, File) :-
   %% 960912 we could drop the File argument to break_line_condition
   %%        but keeping it maintains one clause per breakpoint which
   %%        is easier to handle when adding/removing breakpoints.

   %% 970905 This is now only used to know what lines have breakpoints.
   retractall(break_line_condition(Line, File)),
   assert(break_line_condition(Line, File)),

   update_break_conditions(File).

%% 970509
update_break_conditions(File) :-
   findall(BreakLine, break_line_condition(BreakLine, File), BreakedLines),
   source_debug:define_line_indirection_breaks(File, BreakedLines).


%% 970503 lines with breakpoints for File.
%%        temporary breakpoints flagged by -LineNo
line_breakpoints(File0, Breaks) :-
   %% File0 should allready be an absolute file name 
   %% absolute_file_name(File0,File),
   File = File0,
   findall(Line, line_breakpoint_gen(File,Line), Lines),
   Breaks = Lines.

line_breakpoint_gen(File,Line) :-
   break_line_condition(Line0,File),
   ( 'temporary break'(Line0, File) ->
       Line is -Line0
   ; otherwise ->
       Line = Line0
   ).


%% emacs_remove_line_condition(?Line, ?File)
emacs_remove_line_condition(File0, Line0) :-
   ( nonvar(File0) ->
       absolute_file_name(File0, File),
       ( nonvar(Line0), source_debug:breakable_line(File, Line0, Line) ->
           true
       ; (current_predicate(ale_remove_lc,user:ale_remove_lc(_,_,_)) ->
            user:ale_remove_lc(File0,Line0,Line)
         ;  Line = Line0)
       )
   ; true),
   remove_line_condition_message(Line, File, File0),
   remove_line_condition(File, Line),
   write_goal_skip(on).


remove_line_condition_message(Line, File) :-
   remove_line_condition_message(Line, File, File).


remove_line_condition_message(Line, File, File0) :-
   %% File should be absolute or a variable
   ( var(Line), var(File) ->
       format(user_error, '~N{All breakpoints removed}~n', [])
   ; nonvar(Line) ->                            % nonvar(File) too!
       format(user_error, '10L{REMOVEBREAK ~w "~w"}10R', [Line, File0])% File0!
   ; nonvar(File) ->
       format(user_error, '10L{REMOVEALLBREAKFILE "~w"}10R', [File0])% File0!
   ).
   

remove_line_condition(File, Line) :-
   %% File should be var or an absolute file name
   retractall(break_line_condition(Line, File)),
   retractall('temporary break'(Line, File)),
   update_break_conditions(File).


%% called by act.pl (source_debug)
abolish_line_conditions(File) :-
   retractall(break_line_condition(_, File)),
   retractall('temporary break'(_, File)).

%% break_at(+Line, +File, ?Module, ?Pred)
%% Set a breakpoint at a particular line from emacs
emacs_break_at(File0, Line0) :-
   absolute_file_name(File0, File),
   ( break_at1(Line0, File, BreakLine) ->
       %% 960902 with new scheme we don't really know what predicate
       %%        the line is in. We might add this information some day
       %% format(user_error, '~n{Break on ~w@~w}~n', [Fun/Arity, BreakLine]),
       %% 960918 now mark for special processing
       format(user_error, '10L{BREAK ~w "~w"}10R', [BreakLine, File0]),% File0!
       write_goal_skip(on)
   ; (current_predicate(ale_break_at,user:ale_break_at(_,_)) ->
        user:ale_break_at(File0,Line0)
     ; otherwise ->
         format(user_error, '~n{Could not set breakpoint}~n', []),
         write_goal_skip(on)
     )
   ).

emacs_temporary_break_at(File0, Line0) :-
   absolute_file_name(File0, File),
   write_goal_skip(on),
   ( source_debug:breakable_line(File, Line0, Line) ->
       ( break_line_condition(Line, File) ->
           %% already a breakpoint there, do nothing
           true
       ; break_at1(Line, File, BreakLine) ->    % should never fail here!
           %% -Line signifies temporary break
           format(user_error, '10L{BREAK -~w "~w"}10R', [BreakLine, File0]),% File0!
           %% format(user_error, '~n{Break on line ~w}~n', [BreakLine]),
           record_temporary_break(File, BreakLine)
       )
   ; otherwise ->
       format(user_error, '~n{Could not set breakpoint}~n', [])
   ).

:- dynamic 'temporary break'/2.

record_temporary_break(File, Line) :-
   retractall('temporary break'(Line, File)),
   assert('temporary break'(Line, File)).

remove_temporary_break(File, Line) :-
   ( retract('temporary break'(Line, File)) ->
       %% There was a temporary breakpoint at Line@File, remove it
       %% 970505 send the message to emacs explicitly (used to be
       %%        called redundanly by remove_line_condition)
       remove_line_condition_message(Line, File),
       remove_line_condition(File, Line)
   ; otherwise ->
       true
   ).


break_at1(Line, File, BreakLine) :-
   %% 960912 now done in caller
   %% fix a problem if $HOME is different from ~/ due to
   %% eg remote mounting
   %% absolute_file_name(File0, File),

   ( source_debug:breakable_line(File, Line, BreakLine0) ->
       BreakLine = BreakLine0,
       %% 960902 with new scheme we don't really know what predicate
       %%        the line is in. We might add this information some day
       %% format(user_error, '~n{Break on ~w@~w}~n', [Fun/Arity, BreakLine]),
       %% 960912 do this in caller instead
       %% format(user_error, '~n{Break on line ~w}~n', [BreakLine]),
       add_line_condition(BreakLine, File)
       %% 960912 do this in caller instead
       %% write_goal_skip(on)

   %% 960912 fail to signal failure to caller 
   %% ; otherwise ->
       %% 960912 do this in caller instead
       %% format(user_error, '~n{Could not set breakpoint}~n', []),
       %% write_goal_skip(on)
   ).

                

/* 970509 used to return nearest, now in act.pl (source_debug)
%% Return nearest breakable line or fail if none.
breakable_line(File, Line, BreakLine) :-
   %% File should be from absolute_file_name
   PosGoal =.. [File,BreakLine0],
   between_inclusive_descending(Line, 1, BreakLine0),
   source_debug:'line mapping'(PosGoal, _ClauseNo), !,
   BreakLine = BreakLine0.
*/

between_inclusive_descending(Top, Bot, I) :-
   Top >= Bot,
   I = Top.

between_inclusive_descending(Top, Bot, I) :-
   Top > Bot,
   Top1 is Top-1,
   between_inclusive_descending(Top1, Bot, I).


%% Dump breakable and breaked lines for all konsulted files.
dump_all_breakpoint_info :-
   (
      source_debug:'pos marker goal'(PosGoalHead),
      functor(PosGoalHead,Path,_),
      emacs_file_breakpoint_info1(Path),
      fail
   ; true
   ).



emacs_file_breakpoint_info1(File) :-
   emacs_breakable_lines1(File),
   emacs_breakpoints1(File).

%% 970503 dump breakable lines for a single file
emacs_breakable_lines1(File0) :-
   running_in_emacs, !,
   absolute_file_name(File0,File),
   source_debug:breakable_lines(File, Lines),
   source_debug:compress_line_list(Lines, Cs),
   dump_file_int_list_command(File, Cs, 'BREAKABLELINES').
emacs_breakable_lines1(_File0).



emacs_breakpoints1(File0) :-
   running_in_emacs, !,
   absolute_file_name(File0, File),
   line_breakpoints(File0, Breaks),
   dump_file_int_list_command(File, Breaks, 'BREAKPOINTS').
emacs_breakpoints1(_File0).



/*
dump_line_list(File0, Cs, S) :-
   %% File0 should allready be an absolute file name
   %% absolute_file_name(File0, File),
   File = File0,
   format(S,'~n10L{BREAKABLELINES "~w" (', [File]),
   member(C,Cs),
   write(S,C),
   write(S,' '),
   fail.
dump_line_list(_File0, _Cs, S) :-
   write(S, ')}10R\n').
*/


dump_file_int_list_command(File0, Cs, COMMAND) :-
   %% File0 should allready be an absolute file name
   %% absolute_file_name(File0, File),
   File = File0,
   format(user_error,'~n10L{~w "~w" (', [COMMAND, File]),
   member(C,Cs),
   write(user_error,C),
   write(user_error,' '),
   fail.
dump_file_int_list_command(_File0, _Cs,_COMMAND) :-
   write(user_error, ')}10R\n').

/*
compress_line_list([], []).
compress_line_list(Ls0, [L,Count|LCs]) :- Ls0 = [L|_],
   compress_line_list1(Ls0,Ls, L, Count),
   compress_line_list(Ls, LCs).

compress_line_list1(Ls0,Ls, L, Count) :-
   compress_line_list2(Ls0,Ls, L, 0,Count).

compress_line_list2([],Ls, _L, Count0,Count) :-
   Ls = [],
   Count = Count0.

compress_line_list2(Ls0,Ls, L0, Count0,Count) :- Ls0 = [L0|Ls1], !,
   Count1 is Count0+1,
   L1  is L0+1,
   compress_line_list2(Ls1,Ls, L1, Count1,Count).

compress_line_list2(Ls0,Ls, _L, Count0,Count) :- Ls0 = [_|_],
   Ls = Ls0,
   Count = Count0.


uncompress_line_list([], []).

uncompress_line_list([Count,L0|LCs], Ls) :-
   uncompress_line_list1(Count, L0, Ls,Ls1),
   uncompress_line_list(LCs, Ls1).

uncompress_line_list1(Count, _, Ls0,Ls) :- Count == 0, !,
   Ls = Ls0.

uncompress_line_list1(Count0, L0, [L0|Ls0], Ls) :-
   Count1 is Count0-1,
   L1 is L0+1,
   uncompress_line_list1(Count1, L1, Ls0,Ls).
   
*/

%% %% Dump breakable lines for all files in a konsult arg.
%% write_dumpable_lines(Arg) :- var(Arg), !.       % should not happen

%% write_dumpable_lines(_M:Arg) :- !,
%%    write_dumpable_lines(Arg).

%% write_dumpable_lines(Arg) :- Arg == [], !.

%% write_dumpable_lines([Arg|Args]) :- !,
%%    write_dumpable_lines(Arg),
%%    write_dumpable_lines(Args).

%% write_dumpable_lines(File) :- atom(File), !,
%%    emacs_file_breakpoint_info1(File).

%% write_dumpable_lines(_).
   
   


%% EXTERNAL INTERFACE called from emacs

%%emacs_breakable_lines(File) :-
%%   write_goal_skip(on),
%%   write_dumpable_lines(File).

emacs_file_breakpoint_info(File) :-
   write_goal_skip(on),
   emacs_file_breakpoint_info1(File).


emacs_spy(SpyArg) :-
   write_goal_skip(on),
   preserve_debug(spy(SpyArg)).

emacs_nospy(SpyArg) :-
   write_goal_skip(on),
   nospy(SpyArg).

emacs_nospyall :-
   write_goal_skip(on),
   nospyall.

emacs_leash(LeashArg) :-
   write_goal_skip(on),
   leash(LeashArg).

emacs_compile(Arg) :-
   write_goal_skip(on),
   compile(Arg).

emacs_fcompile(Arg) :-
   write_goal_skip(on),
   fcompile(Arg).

emacs_consult(Arg) :-
   write_goal_skip(on),
   consult(Arg).

emacs_konsult(Arg) :-
   write_goal_skip(on),
   source_debug:konsult(Arg).


emacs_command(HeadTemplate, Goal) :-
   write_goal_skip(on),
   prolog:'$debugger_ancestors'([Current|_CA]),
   Current = a(_Invno, HeadTemplate, _Depth, _Module),
   ( (\+ call(Goal)) -> true; true),
   %% simpleminded workaround for bugs in SICStus buffering scheme (3PL3)
   flush_output(user_output).

emacs_command(HeadTemplate0, InvNo, Goal) :-
   write_goal_skip(on),
   prolog:'$debugger_ancestors'(CA),
   %% ( HeadTemplate0 = _M:HeadTemplate ->
   %%     true
   %% ; otherwise ->
       HeadTemplate = HeadTemplate0
   %% )
   ,
   ( memberchk(a(InvNo, HeadTemplate, _Depth, _Module), CA) ->
       ( (\+ call(Goal)) -> true; true)
   ; otherwise ->
       format(user_error, '~N{Frame ~w is not a current ancestor}~n', [InvNo])
   ),
   %% simpleminded workaround for bugs in SICStus buffering scheme (3PL3)
   flush_output(user_output).


emacs_writeq_toggle :-
   write_goal_skip(on),
   emacs_writeq_enable(Old, Old),
   ( Old = 'on' ->
       New = 'off',
       Msg = '{writeq args: off}\n'
   ; Old = 'off' ->
       New = 'on',
       Msg = '{writeq args: on}\n'
   ),
   emacs_writeq_enable(_Old, New),
   write(user_error, Msg).


emacs_ancestors :-
   write_goal_skip(on),
   %% list(a(+Invno, +Goal, +Depth, +Module))
   prolog:'$debugger_ancestors'(CA),
   %% 10Lreturn value10R
   ( CA = [_] ->
       write(user_error, '10L{ANCESTORS: NONE}10R')
   ; CA = [_|_As] ->
       %% We might as well include the current goal
       do_ancestors1(CA, '10L{ANCESTORS: ')
   ).

emacs_ancestors(N0) :-
   write_goal_skip(on),
   %% list(a(+Invno, +Goal, +Depth, +Module))
   prolog:'$debugger_ancestors'(CA),
   %% 10Lreturn value10R
   ( CA = [_] ->
       write(user_error, '10L{ANCESTORS: NONE}10R')
   ; CA = [_|_As] ->
       %% We might as well include the current goal
       N is N0+1,
       do_ancestors1(CA, N, '10L{ANCESTORS: ')
   ).

do_ancestors1([], _N0,_Prefix) :-
   write(user_error, '}10R'),
   nl(user_error).

do_ancestors1([a(Invno, Goal, _Depth, Module)|CA], N0, Prefix) :-
   ( N0 == 0 ->
       do_ancestors1([], N0,Prefix)
   ; otherwise ->
       N1 is N0-1,
       functor(Goal, F, A),
       format(user_error, '~w[~d, ~q, ~q, ~d]', [Prefix, Invno, Module, F, A]),
       do_ancestors1(CA, N1, ', ')
   ).


do_ancestors1([], _Prefix) :-
   write(user_error, '}10R'),
   nl(user_error).

do_ancestors1([a(Invno, Goal, _Depth, Module)|CA], Prefix) :-
   functor(Goal, F, A),
   format(user_error, '~w[~d, ~q, ~q, ~d]', [Prefix, Invno, Module, F, A]),
   do_ancestors1(CA, ', ').


informational_message_hook(loaded(How,Path,Module,_,_)) :- !,
   ( How = 'loaded' ->
       non_konsult_hook(Path, Module)
   ; How = 'compiled' ->
       non_konsult_hook(Path, Module)
   ; How = 'consulted' ->
       non_konsult_hook(Path, Module)
   ; How = 'fcompiled' ->
       true
   ),
   fail.
   
/*
(progn
   (insert "\n{file: \'/home/groucho/csd/perm/prolog/source_debug/kaka.pl\' line: 17}{SYNTAX ERROR: in lines 14-18, [operator,expected,after,expression] (token 9)}\n")
   (goto-char (point-min))
   (re-search-forward " \\[\\(\\([A-Z_][A-Za-z0-9_]*,?\\)+\\)\\] - singleton variables in" 1599 t))
*/


portray_error(Message) :-
   with_unbuffered_user_error(portray_error1(Message)).


portray_error1(Message) :-
   Message = syntax_error(Goal,between(AL,BL),Msg,Tokens,AfterError),
   prefix_file_info1(FormatString, Args), !,
   %% write(user_error, Message), nl(user_error),
   %% We have a syntax error and we know where it originated.
   length(Tokens, Length),
   BeforeError is Length-AfterError,
   format(user_error, FormatString, Args),
   ( syntax_error_goal_summary(Goal, GoalSummary) -> % e user:foo/3
       format(user_error, '{SYNTAX ERROR: ~w - in lines ~d-~d, ',
              [GoalSummary, AL,BL]),
       write_syntax_msg(Msg),
       format(user_error, '(token ~w)}~n', [BeforeError])
   
   ; otherwise ->
       format(user_error, '{SYNTAX ERROR: in lines ~d-~d, ', [AL,BL]),
       write_syntax_msg(Msg),
       format(user_error, '(token ~w)}~n', [BeforeError])
   ).

portray_error1(Msg) :-
   Msg = existence_error(Goal,ArgNo,ObjType,Culprit,Message),
   ( ( nonvar(Message),
       Message = past_end_of_stream ) ->
       fail                              % don't handle it here.
   ; ArgNo = 0 ->
       !,                                       % comit to this message
       format(user_error,
              '{EXISTENCE ERROR: ~q ~q does not exist}~n', % DON'T write out goal
              [ObjType,Culprit])
   ; otherwise ->
       !,                                       % comit to this message
       format(user_error,
              '{EXISTENCE ERROR: ~q - arg ~d: ~q ~q does not exist}~n', % Que?
              [Goal,ArgNo,ObjType,Culprit])
   ).

write_syntax_msg([]).

write_syntax_msg([W|Ws]) :-
   ( W == '{' ->                               % dont mess up {} matching
       write_syntax_msg([left,brace|Ws])
   ; W == '}' ->
       write_syntax_msg([right,brace|Ws])
   ; otherwise ->
       write(user_error, W),
       write(user_error, ' '),
       write_syntax_msg(Ws)
   ).


syntax_error_goal_summary(Goal, _GoalSummary) :- Goal = 0, !,
   fail.

syntax_error_goal_summary(Module:Goal, GoalSummary) :- !,
   GoalSummary = Module:GoalSummary1,
   syntax_error_goal_summary(Goal, GoalSummary1).

syntax_error_goal_summary(Goal, GoalSummary) :-
   functor(Goal, F, A),
   GoalSummary = F/A.


prefix_file_info_then_fail :-
   prefix_file_info,
   fail.                                        % do normal processing


prefix_file_info :-
   prefix_file_info1(FormatString, Args),       % might fail
   format(user_error, FormatString, Args).


prefix_file_info1(FormatString, Args) :-
   running_in_emacs,                            % 960902 not the ideal place
   %% order is important, this way load/consult can call konsult
   %% but not the other way around
   ( (  %% source_debug module might not be loaded (Yet! while loading this file)
        current_predicate(konsult_context, source_debug:konsult_context(_,_)),
        source_debug:konsult_context(file, Path)
     ) ->
       FormatString = '~N{file: \'~w\'}',
       Args = [Path]
   ; prolog_load_context(file, Path) ->
       prolog_load_context(term_position,
                           '$stream_position'(_CC,LC,_LP,_Seek)),
       %% bug in SICStus 3PL2, 
       %% prolog_load_context returns multiple solutions if doing
       %% recursive file operations, ie compiling from a consulted
       %% file. 
       !,
       FormatString = '~N{file: \'~w\' line: ~w}',
       Args = [Path, LC]
   ).
   
/* to 960913
prefix_file_info :-
   running_in_emacs,                            % 960902 not the ideal place
   %% order is important, this way load/consult can call konsult
   %% but not the other way around
   ( (  %% source_debug module might not be loaded (960902 Yeah Right...)
        current_predicate(konsult_context, source_debug:konsult_context(_,_)),
        source_debug:konsult_context(file, Path)
     ) ->
       format(user_error, '~N{file: \'~w\'}', [Path])
   ; prolog_load_context(file, Path) ->
       prolog_load_context(term_position,
                           '$stream_position'(_CC,LC,_LP,_Seek)),
       %% bug in SICStus 3PL2, 
       %% prolog_load_context returns multiple solutions if doing
       %% recursive file operations, ie compiling from a consulted
       %% file. 
       !,
       format(user_error, '~N{file: \'~w\' line: ~w}', [Path, LC])
   ).
*/


/* To use:
with_unbuffered_user_error(Goal)

or:

kludge_stream_open(user_error, Cookie),
writeq(user_error, ...), ....
%% everything written to user_error will appear on stdout
kludge_stream_close(Cookie).

Note: ensure that kludge_stream_close is ALWAYS called!

*/
%% at top of file
%% :- meta_predicate with_unbuffered_user_error(:).

with_unbuffered_user_error(Goal) :-
   kludge_stream_open(Cookie),
   ( on_exception(E,
                  Goal,
                 ( kludge_stream_close(Cookie),
                   raise_exception(E)
                 )),
     !,
     kludge_stream_close(Cookie)
   ; kludge_stream_close(Cookie),
     fail
   ).


%% Consider exploiting the fact that we will be called in a context
%% where unknown predicates fail. It might even be faster to ensure
%% this here than to use current_predicate.
:- dynamic kludge_stderr/1.
kludge_stream_open(Cookie) :-
   %% fail,
   ( kludge_stderr(StdErr) ->
       true
   ; current_predicate(buffer_mode, buffer_mode:buffer_mode(_,_)) ->
       open(2, write, StdErr),
       ( true ->
           buffer_mode:buffer_mode(StdErr, blockbuff)
       ; true
       ),
       assert(kludge_stderr(StdErr))
   ),!,
   Cookie = open2_kludge(StdErr,Old),
   prolog_flag(user_error,Old,StdErr).
kludge_stream_open(kludge(user_error)).         % added 970822 if no buffer_mode

%%%% 960923 now use open(2,... instead)
%%%% compatible defs, no-ops unless module buffer_mode is loaded
kludge_stream_open(StreamName, Cookie) :-
   ( current_predicate(buffer_mode, buffer_mode:buffer_mode(_,_)) ->
       kludge_stream_open_internal_raw(StreamName, Cookie)
   ; otherwise ->
       Cookie = kludge(StreamName)
   ),
   !.

kludge_stream_open(StreamName, Cookie) :-
   StreamName = user_error,                     % should drop this arg.
   kludge_stream_open(Cookie).

kludge_stream_open_internal_raw(StreamName, Cookie) :-
   prolog_flag(StreamName, S),
   buffer_mode(S, blockbuff),
   Cookie = raw_cookie(StreamName, S).


kludge_stream_close(Cookie) :- var(Cookie), !,
   fail.

kludge_stream_close(open2_kludge(S,Old)) :-
   flush_output(S),
   prolog_flag(user_error, _, Old).

kludge_stream_close(kludge(StreamName)) :-
   flush_output(StreamName).

kludge_stream_close(raw_cookie(StreamName, S)) :- !,
   flush_output(StreamName),
   buffer_mode(S, nobuff).

%% Hooks added at end to ensure that everything is defined while loading THIS file
/*
:- multifile user:term_expansion/2.
user:term_expansion(end_of_file,_) :-
   term_expansion_hook.                         % will fail!!
*/


%%% prefix errors and warnings with file and line context
%%% This also supports the source_debug package if loaded
:- multifile user:portray_message/2.            % !!
user:portray_message(error, Message) :-
   portray_error(Message), !.

user:portray_message(warning, _Message) :-
   prefix_file_info_then_fail.

user:portray_message(informational, Message) :-
   informational_message_hook(Message), !.


/* to 960902
%% Set a breakpoint at a particular line from emacs
%% Note that Pred can be a var. We should consider using pred only as
%% a hint so that even a bad guess in emacs does nothing bad except
%% slow things down a bit.
%% Same goes for Module
%% break_at(+Line, +File, ?Module, ?Pred)
break_at(Line0, File0, Module, _Pred) :-
   %% fix a problem if $HOME is different from ~/ due to
   %% eg remote mounting
   absolute_file_name(File0, File),
   %% currently Pred is assumed to be a variable
   ( nonvar(Module) ->
       Ms = [Module]
   ; otherwise ->
       Ms = [user|Ms0],                        % try user first,
                                               % ignore prolog
       ( setof(CM,
               (
                   current_module(CM),
                   CM \== prolog,
                   CM \== user,
                   %% We adhere to the policy that a module should consist of only one file
                   %% This is for speed only, the call to
                   %% current_module/2 can simply be commented out.
                   %% This is especailly bogus since konsulted files
                   %% will not set the module->file mapping correctly so
                   %% this will only be right if the file was previously
                   %% consulted or loaded. Anyway, module support is
                   %% not-so-good in the source debugger at large.
                   current_module(CM,File)
               ), 
               Ms0) -> true ; Ms0 = [] )
   ),
   %% First try for exact match on line number
   ( ( dbg_member(M, Ms),
       %% 951129 all konsulted are dynamic
       predicate_property(M:H, dynamic),
       %% current_predicate(Pred, M:H),
       functor(H, F, A),
       line_at_or_before_in_pred(M, F, A, File, Line0, Line0) ) ->
       %% Found pos/2 at exact line
       Break = Line0,
       Fun = F,
       Arity = A
   ; otherwise ->                               % slower, try to find preceeding line to break at
       format(user_error, '~n{Looking for preceeding line}~n', []),
       findall(FA-Lines,
               (
                   dbg_member(M, Ms),
                   %% 951129 all konsulted are dynamic
                   predicate_property(M:H, dynamic),
                   %% current_predicate(Pred, M:H),

                   functor(H, F, A),
                   FA = F/A,
                   setof(BreakLine, line_at_or_before_in_pred(M, F, A, File, Line0, BreakLine), Lines)
                ),
               PredLines),
        %% format(user_error, '~nFound ~w for line ~w in ~w~n', [PredLines, Line0, Head]),
        choose_appropriate_line_breakpoint(PredLines, Line0, Fun/Arity, Break)
   ), !,
   format(user_error, '~n{Break on ~w@~w}~n', [Fun/Arity, Break]),

   add_line_condition(Break, File),
   write_goal_skip(on).

break_at(_Line0, _File, _Module, _Pred) :-
   format(user_error, '~n{Could not set breakpoint}~n', []),
   write_goal_skip(on).


line_at_or_before_in_pred(Module, F, A, File, Line, BreakLine) :-
   functor(H, F, A),
   safe_clause(Module:H, Body),
   line_at_or_before_in_body(Body, File, Line, BreakLine).


choose_appropriate_line_breakpoint(PredLines, Line, FA, Break) :-
   FA0 = none,
   Break0 = 0,
   choose_appropriate_line_breakpoint(PredLines, Line, FA0, Break0, FA, Break).


choose_appropriate_line_breakpoint([], _Line, FA0, Break0, FA, Break) :-
   FA = FA0,
   Break = Break0.

choose_appropriate_line_breakpoint([FA1-Lines|PredLines], Line, FA0, Break0, FA, Break) :-
   ( dbg_member(Line, Lines) ->
       Break = Line,
       FA = FA1
   ; otherwise ->
       find_int_preceeding(Lines, Line, Break1),
       ( Break1 > Break0 ->
           FA2 = FA1,
           Break2 = Break1
       ; otherwise ->
           FA2 = FA0,
           Break2 = Break0
       ),
       choose_appropriate_line_breakpoint(PredLines, Line, FA2, Break2, FA, Break)
   ).


safe_clause(MH,Body) :-
   on_exception(_Expt,                          % access errors for non-konsulted preds
      clause(MH, Body),
      fail
   ).


line_at_or_before_in_body(Body, _File, _Line, _BreakLine) :- var(Body), !, fail.

line_at_or_before_in_body((A,B), File, Line, BreakLine) :-
   (
       line_at_or_before_in_body(A, File, Line, BreakLine)
   ;
       line_at_or_before_in_body(B, File, Line, BreakLine)
   ).

line_at_or_before_in_body((A;B), File, Line, BreakLine) :-
   (
       line_at_or_before_in_body(A, File, Line, BreakLine)
   ;
       line_at_or_before_in_body(B, File, Line, BreakLine)
   ).

line_at_or_before_in_body((A->B), File, Line, BreakLine) :-
   (
       line_at_or_before_in_body(A, File, Line, BreakLine)
   ;
       line_at_or_before_in_body(B, File, Line, BreakLine)
   ).

line_at_or_before_in_body(user:'$ pos'(Pos, File), File, Line, BreakLine) :-
   Pos =< Line,
   Pos > Line-10,                               % arbitrary limit
   BreakLine = Pos.

%% 960222 module will not appear if user code, kludge it
%%       (will go wrong if other module)
line_at_or_before_in_body('$ pos'(Pos, File), File, Line, BreakLine) :-
   Pos =< Line,
   Pos > Line-10,                               % arbitrary limit
   BreakLine = Pos.


find_int_preceeding([I0|Is], I, Found) :-
   find_int_preceeding(Is, I, I0, Found).


find_int_preceeding([], _I, Found0, Found) :-
   Found = Found0.

find_int_preceeding([I0|Is], I, Found0, Found) :-
   ( I0 > I ->
       Found = Found0
   ; otherwise ->
       find_int_preceeding(Is, I, I0, Found)
   ).

*/

/* 970509 Now done by modifying the line marker goal
add_line_condition(Line, File) :-
   without_user_error(preserve_debug(add_line_condition1(Line, File))).


add_line_condition1(Line, File) :-
   %% nospy(user:'$ pos'/2),
   source_debug:src_pos_goal(SomeLine, File, PosHead),
   nospy(user:File/1),
   %% 960912 we could drop the File argument to break_line_condition
   %%        but keeping it maintains one clause per breakpoint which
   %%        is easier to handle when adding/removing breakpoints.
   spypoint_condition(PosHead,                  % user:File(_)
                      call,
                      %% don't forget to change this if module name changes
                      emacs:break_line_condition(SomeLine, File)),
   retractall(break_line_condition(Line, File)),
   assert(break_line_condition(Line, File)).

remove_line_condition(File, Line) :-
   %% File should be var or an absolute file name

   %% 970505 removed as it should be done in caller and not always anyway.
   %% ( retract(break_line_condition(Line, File)) ->
   %%     format(user_error, '10L{REMOVEBREAK ~w "~w"}10R', [Line, File])
   %% ; true
   %% ),
   retractall(break_line_condition(Line, File)),
   retractall('temporary break'(Line, File)),
   ( break_line_condition(_, _) ->
       true
   ; otherwise ->                               % no line breaks
       %% nospy(user:'$ pos'/2)                % for speed
       without_user_error(nospy(user:File/1))
   ).

*/
/* To 9?0913
user:portray_message(warning, Message) :- %% !,
   portray_warning(Message), !.
   %% prefix_file_info_then_fail.

user:portray_message(error, _Message) :- %% !,
   prefix_file_info_then_fail.
*/

